//
// Pixelstick Software
//
// Author: Alex Merz
// Datum: 31.03.2018
//
// Bild: Das Bild muss folgendes Format aufweisen:
//       -144-Pixel auf x-Pixel oder 288-Pixel auf x-Pixel
//       -BMP-Format
//       -24Bit
//       -Bild um 90-Grad Gegenuhrzeigersinn gedreht
//
//
////////////////////////////////////////////////////////////////////////////////Includes
#include <Adafruit_NeoPixel.h> //Neopixel
#include <SD.h> //SD
#include <SPI.h> //SPI Interface


////////////////////////////////////////////////////////////////////////////////Pinbelegung
//Neopixel
#define PinNeopixel 2 //Mega 2

//Trigger
#define PinTrigger 36 //Mega 36
#define PinEndlosTrigger 38 //Mega 38

//Potentiometer
#define PinPotiHelligkeit A0 //Mega A0
#define PinPotiBildZeit A1 //Mega A1

//Wahlschalter
#define PinWahl1 22 //Mega 22
#define PinWahl2 24 //Mega 24
#define PinWahl3 26 //Mega 26
#define PinWahl4 28 //Mega 28

//SD
#define PinSDMiso 50 //Mega 50
#define PinSDMosi 51 //Mega 51
#define PinSDSck 52 //Mega 52
#define PinSDCs 4 //Mega 53


////////////////////////////////////////////////////////////////////////////////Variabelndeklaration
//SD
File root;
File dataFile;
String m_CurrentFilename = "";
int m_FileIndex = 0;
int m_NumberOfFiles = 0;
String m_FileNames[200];

//Neopixel
int r = 0; //Variable für Rot-Anteil
int g = 0; //Variable für Grün-Anteil
int b = 0; //Variable für Blau-Anteil

int Helligkeit = 100;
int frameDelay = 0;
int Wahlschaltercode = 0000;


////////////////////////////////////////////////////////////////////////////////Konstantendeklaration
#define Neopixellaenge 288


////////////////////////////////////////////////////////////////////////////////Funktiondeklaration
Adafruit_NeoPixel Ledstrip = Adafruit_NeoPixel(Neopixellaenge, PinNeopixel, NEO_GRB + NEO_KHZ800); //Neopixel
void SDInit(void);
void setupLEDs(void);
void BildAusgeben(void);
void Bildwahl(void);
void PotiEinlesen(void);


////////////////////////////////////////////////////////////////////////////////Initialisierungs-Funktion
void setup() 
{
  //Serial-Port
  Serial.begin(9600);  

  //Trigger
  pinMode(PinTrigger, INPUT_PULLUP);
  pinMode(PinEndlosTrigger, INPUT_PULLUP);

  //Wahlschalter
  pinMode(PinWahl1, INPUT_PULLUP);
  pinMode(PinWahl2, INPUT_PULLUP);
  pinMode(PinWahl3, INPUT_PULLUP);
  pinMode(PinWahl4, INPUT_PULLUP);

  //SD-Karte
  SDInit();

  //Neopixel
  setupLEDs();

  Serial.println("Initialisierung abgeschlossen");
  Serial.println("System bereit fuer Bildausgabe");
}


////////////////////////////////////////////////////////////////////////////////Hauptfunktion
void loop() 
{
  //Bild ausgeben
  if((digitalRead(PinTrigger) == LOW)||(digitalRead(PinEndlosTrigger) == LOW))
  { 
    BildAusgeben();
  }

  //Potentiometer einlesen
  PotiEinlesen();

  //Bildwahl einlesen
  Bildwahl();
}


////////////////////////////////////////////////////////////////////////////////Funtkionen
////////////////////////////////////////////////////////////////////////////////BildAusgeben
void BildAusgeben()
{
  Serial.println("Bild auf Neopixelstrip ausgeben...");

  do
  {
    SendFile(m_CurrentFilename);
    ClearStrip(0);
  }while(digitalRead(PinEndlosTrigger) == LOW);

  Serial.println("Bild ausgegeben");
  Serial.println("System bereit fuer Bildausgabe");
}


////////////////////////////////////////////////////////////////////////////////PotiEinlesen
void PotiEinlesen()
{
  //Serial.println("Poti einlesen und skalieren");
  //Serial.println(analogRead(PinPotiHelligkeit));
  //Serial.println(analogRead(PinPotiBildZeit));
  //delay(100);

  Helligkeit = map(analogRead(PinPotiHelligkeit), 0, 1020, 0, 100);

  frameDelay = map(analogRead(PinPotiBildZeit), 0, 1020, 0, 100);
  
  //Serial.println(frameDelay);
  //Serial.println(Helligkeit);
  //delay(100);
}


////////////////////////////////////////////////////////////////////////////////Bildwahl
void Bildwahl()
{
  //Serial.println("Bildwahl");
  //Serial.println("Wahlschalter auslesen");
  //Serial.print(digitalRead(PinWahl1));
  //Serial.print(digitalRead(PinWahl2));
  //Serial.print(digitalRead(PinWahl3));
  //Serial.println(digitalRead(PinWahl4));
  //delay(1000);

  Wahlschaltercode = digitalRead(PinWahl1)*1000 + digitalRead(PinWahl2)*100 + digitalRead(PinWahl3)*10 + digitalRead(PinWahl4);
  
  //Serial.println(Wahlschaltercode);
  //delay(1000);

  switch( Wahlschaltercode)
  {
    //Position 0
    case 1111:
      m_FileIndex = 0; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0     
      DisplayCurrentFilename();
    break;

    
    //Position 1
    case 1110:
      if(m_NumberOfFiles < 2) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 1; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;   


    //Position 2
    case 1101:
      if(m_NumberOfFiles < 3) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 2; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 3
    case 1100:
      if(m_NumberOfFiles < 4) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 3; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;  


    //Position 4
    case 1011:
      if(m_NumberOfFiles < 5) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 4; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 5
    case 1010:
      if(m_NumberOfFiles < 6) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 5; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;


    //Position 6
    case 1001:
      if(m_NumberOfFiles < 7) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 6; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 7
    case 1000:
      if(m_NumberOfFiles < 8) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 7; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 8
    case 111:
      if(m_NumberOfFiles < 9) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 8; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 9
    case 110:
      if(m_NumberOfFiles < 10) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 9; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 10 (A)
    case 101:
      if(m_NumberOfFiles < 11) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 10; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break; 


    //Position 11 (B)
    case 100:
      if(m_NumberOfFiles < 12) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 11; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;


    //Position 12 (C)
    case 11:
      if(m_NumberOfFiles < 13) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 12; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;


    //Position 13 (D)
    case 10:
      if(m_NumberOfFiles < 14) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 13; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;


    //Position 14 (E)
    case 1:
      if(m_NumberOfFiles < 15) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 14; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;


    //Position 15 (F)
    case 0:
      if(m_NumberOfFiles < 16) //m_NumberOfFiles = Anzahl Bilder auf SD
      {
        m_FileIndex = 0;
      }
      else
      {
        m_FileIndex = 15; //m_FileIndex = Position im Array, für Bild Wiedergabe. Achtung beginnt bei 0
      }
      
      DisplayCurrentFilename();
    break;                                                 
  }
}


////////////////////////////////////////////////////////////////////////////////setupLEDs
void setupLEDs() 
{
  Ledstrip.begin();
  Ledstrip.show();
}


////////////////////////////////////////////////////////////////////////////////SDInit
void SDInit()
{
  pinMode(PinSDCs, OUTPUT);

  while (!SD.begin(PinSDCs)) 
  {
    Serial.println("SD Initialisierung fehlgeschlagen");
  }

  Serial.println("SD Initialisierung erfolgreich");
  root = SD.open("/");
  Serial.println("Files wurden gescannt");
  
  GetFileNamesFromSD(root);
  isort(m_FileNames, m_NumberOfFiles);
  m_CurrentFilename = m_FileNames[0];
  DisplayCurrentFilename();    
}


////////////////////////////////////////////////////////////////////////////////SendFile
void SendFile(String Filename) 
{
  char temp[14];
  Filename.toCharArray(temp, 14);
  dataFile = SD.open(temp);

  //Wenn File da ist, sende es an Neopixel-Strip
  if (dataFile) 
  {
    ReadTheFile();
    dataFile.close();
  }
  else 
  {
    Serial.println("Fehler beim einlesen des Files");
    SDInit();
    return;
  }
}


////////////////////////////////////////////////////////////////////////////////DisplayCurrentFilename
void DisplayCurrentFilename() 
{
  m_CurrentFilename = m_FileNames[m_FileIndex];
  //Serial.print("Filename: ");
  //Serial.println(m_CurrentFilename);
}


////////////////////////////////////////////////////////////////////////////////GetFileNamesFromSD
void GetFileNamesFromSD(File dir) 
{
  int fileCount = 0;
  String CurrentFilename = "";
  while (1) 
  {
    File entry =  dir.openNextFile();
    if (! entry) 
    {
      // no more files
      m_NumberOfFiles = fileCount;
      entry.close();
      break;
    }
    else 
    {
      if (entry.isDirectory()) {
        //GetNextFileName(root);
      }
      else 
      {
        CurrentFilename = entry.name();
        if (CurrentFilename.endsWith(".bmp") || CurrentFilename.endsWith(".BMP") ) 
        { //find files with our extension only
          m_FileNames[fileCount] = entry.name();
          fileCount++;
        }
      }
    }
    entry.close();
  }
}


////////////////////////////////////////////////////////////////////////////////latchanddelay
void latchanddelay(int dur) 
{
  Ledstrip.show();
  delay(dur);
}


////////////////////////////////////////////////////////////////////////////////ClearStrip
void ClearStrip(int duration) 
{
  int x;
  for (x = 0; x < Neopixellaenge; x++) 
  {
    Ledstrip.setPixelColor(x, 0);
  }
  Ledstrip.show();
}


////////////////////////////////////////////////////////////////////////////////readLong
uint32_t readLong() 
{
  uint32_t retValue;
  byte incomingbyte;

  incomingbyte = readByte();
  retValue = (uint32_t)((byte)incomingbyte);

  incomingbyte = readByte();
  retValue += (uint32_t)((byte)incomingbyte) << 8;

  incomingbyte = readByte();
  retValue += (uint32_t)((byte)incomingbyte) << 16;

  incomingbyte = readByte();
  retValue += (uint32_t)((byte)incomingbyte) << 24;

  return retValue;
}


////////////////////////////////////////////////////////////////////////////////readInt
uint16_t readInt() 
{
  byte incomingbyte;
  uint16_t retValue;

  incomingbyte = readByte();
  retValue += (uint16_t)((byte)incomingbyte);

  incomingbyte = readByte();
  retValue += (uint16_t)((byte)incomingbyte) << 8;

  return retValue;
}


////////////////////////////////////////////////////////////////////////////////readByte
int readByte() 
{
  int retbyte = -1;
  while (retbyte < 0) 
  {
    retbyte = dataFile.read();
  }
  return retbyte;
}


////////////////////////////////////////////////////////////////////////////////getRGBwithGamma
void getRGBwithGamma() 
{
  g = gamma(readByte()) / (101 - Helligkeit);
  b = gamma(readByte()) / (101 - Helligkeit);
  r = gamma(readByte()) / (101 - Helligkeit);
}


////////////////////////////////////////////////////////////////////////////////ReadTheFile
void ReadTheFile() 
{
#define MYBMP_BF_TYPE           0x4D42
#define MYBMP_BF_OFF_BITS       54
#define MYBMP_BI_SIZE           40
#define MYBMP_BI_RGB            0L
#define MYBMP_BI_RLE8           1L
#define MYBMP_BI_RLE4           2L
#define MYBMP_BI_BITFIELDS      3L

  uint16_t bmpType = readInt();
  uint32_t bmpSize = readLong();
  uint16_t bmpReserved1 = readInt();
  uint16_t bmpReserved2 = readInt();
  uint32_t bmpOffBits = readLong();
  bmpOffBits = 54;

  /* Check file header */
  if (bmpType != MYBMP_BF_TYPE || bmpOffBits != MYBMP_BF_OFF_BITS) 
  {
    Serial.println("Es ist kein Bitmap");
    return;
  }

  /* Read info header */
  uint32_t imgSize = readLong();
  uint32_t imgWidth = readLong();
  uint32_t imgHeight = readLong();
  uint16_t imgPlanes = readInt();
  uint16_t imgBitCount = readInt();
  uint32_t imgCompression = readLong();
  uint32_t imgSizeImage = readLong();
  uint32_t imgXPelsPerMeter = readLong();
  uint32_t imgYPelsPerMeter = readLong();
  uint32_t imgClrUsed = readLong();
  uint32_t imgClrImportant = readLong();

  /* Check info header */
  if ( imgSize != MYBMP_BI_SIZE || imgWidth <= 0 ||
       imgHeight <= 0 || imgPlanes != 1 ||
       imgBitCount != 24 || imgCompression != MYBMP_BI_RGB ||
       imgSizeImage == 0 )
  {
    Serial.println("Nicht unterstuetzte Bitmap. Benutze 24Bit");   
    return;
  }
  
  int displayWidth = imgWidth;
  if (imgWidth > Neopixellaenge) 
  {
    displayWidth = Neopixellaenge;           //only display the number of led's we have
  }
  
  /* compute the line length */
  uint32_t lineLength = imgWidth * 3;
  if ((lineLength % 4) != 0)
    lineLength = (lineLength / 4 + 1) * 4;
    
  // Note:
  // The x,r,b,g sequence below might need to be changed if your strip is displaying
  // incorrect colors.  Some strips use an x,r,b,g sequence and some use x,r,g,b
  // Change the order if needed to make the colors correct.

  for (int y = imgHeight; y > 0; y--) 
  {
    int bufpos = 0;
    for (int x = 0; x < displayWidth; x++) 
    {
      uint32_t offset = (MYBMP_BF_OFF_BITS + (((y - 1) * lineLength) + (x * 3))) ;
      dataFile.seek(offset);

      getRGBwithGamma();

      Ledstrip.setPixelColor(x, r, b, g);

    }
    latchanddelay(frameDelay);
  }
}


////////////////////////////////////////////////////////////////////////////////isort
// Sort the filenames in alphabetical order
void isort(String *filenames, int n) 
{
  for (int i = 1; i < n; ++i) 
  {
    String j = filenames[i];
    int k;
    for (k = i - 1; (k >= 0) && (j < filenames[k]); k--) 
    {
      filenames[k + 1] = filenames[k];
    }
    filenames[k + 1] = j;
  }
}


////////////////////////////////////////////////////////////////////////////////gammaTable
PROGMEM const unsigned char gammaTable[]  = {
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,
  2,  2,  2,  2,  2,  3,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,
  4,  4,  4,  4,  5,  5,  5,  5,  5,  6,  6,  6,  6,  6,  7,  7,
  7,  7,  7,  8,  8,  8,  8,  9,  9,  9,  9, 10, 10, 10, 10, 11,
  11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 15, 15, 16, 16,
  16, 17, 17, 17, 18, 18, 18, 19, 19, 20, 20, 21, 21, 21, 22, 22,
  23, 23, 24, 24, 24, 25, 25, 26, 26, 27, 27, 28, 28, 29, 29, 30,
  30, 31, 32, 32, 33, 33, 34, 34, 35, 35, 36, 37, 37, 38, 38, 39,
  40, 40, 41, 41, 42, 43, 43, 44, 45, 45, 46, 47, 47, 48, 49, 50,
  50, 51, 52, 52, 53, 54, 55, 55, 56, 57, 58, 58, 59, 60, 61, 62,
  62, 63, 64, 65, 66, 67, 67, 68, 69, 70, 71, 72, 73, 74, 74, 75,
  76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91,
  92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 104, 105, 106, 107, 108,
  109, 110, 111, 113, 114, 115, 116, 117, 118, 120, 121, 122, 123, 125, 126, 127
};


////////////////////////////////////////////////////////////////////////////////gamma
inline byte gamma(int x) 
{
  return pgm_read_byte(&gammaTable[x]);
}


